/* finetd, an Internet Superserver
 *
 * Supports IPv4, IPv6 and UNIX domain sockets
 *
 ** Kris Katterjohn <katterjohn@gmail.com>
 ** http://finetd.sourceforge.net
 *
 * Copyright (c) 2008 Kris Katterjohn
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/* $Id: builtin.c 12 2008-02-23 00:06:01Z kjak $ */
/* https://finetd.svn.sourceforge.net/svnroot/finetd */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <assert.h>
#include <sys/socket.h>

#include "finetd.h"

/* Is the service with this protocol and port number built in? */
int is_builtin(unsigned short port, int proto)
{
	/* No built-in UNIX services yet */
	if (proto != IPPROTO_TCP && proto != IPPROTO_UDP)
		return 0;

	switch (port) {
	case IPPORT_ECHO:
	case IPPORT_DISCARD:
	case IPPORT_DAYTIME:
	case IPPORT_CHARGEN:
		return 1;
	default:
		break;
	}

	return 0;
}

static void _tcp_echo(void)
{
	char buf[256];
	int c;

	for (;;) {
		c = read(0, buf, sizeof buf);

		if (c <= 0)
			break;

		if (write(1, buf, c) <= 0)
			break;
	}
}

static void _udp_echo(void)
{
	char buf[256];
	int c;
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof ss;

	c = recvfrom(0, buf, sizeof buf, 0, (SA *) &ss, &sslen);

	if (c <= 0)
		return;

	if (sslen < sizeof(SA) || sslen > sizeof ss)
		return;

	sendto(1, buf, c, 0, (SA *) &ss, sslen);
}

/* Writes back anything that's read */
void builtin_echo(int proto)
{
	switch (proto) {
	case IPPROTO_TCP:
		_tcp_echo();
		break;
	case IPPROTO_UDP:
		_udp_echo();
		break;
	default:
		assert(0);
	}
}

static void _tcp_discard(void)
{
	char buf[256];

	while (read(0, buf, sizeof buf) > 0);
}

static void _udp_discard(void)
{
	char buf[256];

	read(0, buf, sizeof buf);
}

/* Reads; doesn't write */
void builtin_discard(int proto)
{
	switch (proto) {
	case IPPROTO_TCP:
		_tcp_discard();
		break;
	case IPPROTO_UDP:
		_udp_discard();
		break;
	default:
		assert(0);
	}
}

static void _tcp_daytime(void)
{
	time_t t = time(NULL);
	char *c = ctime(&t);

	write(1, c, strlen(c));
}

static void _udp_daytime(void)
{
	time_t t = time(NULL);
	char *c = ctime(&t);
	char buf[1];
	int ret;
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof ss;

	ret = recvfrom(0, buf, sizeof buf, 0, (SA *) &ss, &sslen);

	if (ret < 0)
		return;

	if (sslen < sizeof(SA) || sslen > sizeof ss)
		return;

	sendto(1, c, strlen(c), 0, (SA *) &ss, sslen);
}

/* Returns time in ASCII format */
void builtin_daytime(int proto)
{
	switch (proto) {
	case IPPROTO_TCP:
		_tcp_daytime();
		break;
	case IPPROTO_UDP:
		_udp_daytime();
		break;
	default:
		assert(0);
	}
}

/* TODO: shift this like in the RFC */
static char *cgenstr = "ABCDEFGHIJKLMNOPQRSTUWXYZ0123456789`~!@#$%^&*()-_=+";

static void _tcp_chargen(void)
{
	while (write(1, cgenstr, strlen(cgenstr)) > 0);
}

static void _udp_chargen(void)
{
	char buf[1];
	int ret;
	struct sockaddr_storage ss;
	socklen_t sslen = sizeof ss;

	ret = recvfrom(0, buf, sizeof buf, 0, (SA *) &ss, &sslen);

	if (ret < 0)
		return;

	if (sslen < sizeof(SA) || sslen > sizeof ss)
		return;

	sendto(1, cgenstr, strlen(cgenstr), 0, (SA *) &ss, sslen);
}

/* Continually writes the string */
void builtin_chargen(int proto)
{
	switch (proto) {
	case IPPROTO_TCP:
		_tcp_chargen();
		break;
	case IPPROTO_UDP:
		_udp_chargen();
		break;
	default:
		assert(0);
	}
}

