/* finetd, an Internet Superserver
 *
 * Supports IPv4, IPv6 and UNIX domain sockets
 *
 ** Kris Katterjohn <katterjohn@gmail.com>
 ** http://finetd.sourceforge.net
 *
 * Copyright (c) 2008 Kris Katterjohn
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 *
 ****
 * 01/05/2008
 */

/* $Id: finetd.c 12 2008-02-23 00:06:01Z kjak $ */
/* https://finetd.svn.sourceforge.net/svnroot/finetd */

#include <unistd.h>
#include <sys/select.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <setjmp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/socket.h>
#include <sys/wait.h>

#include "finetd.h"
#include "builtin.h"
#include "error.h"
#include "parseconf.h"
#include "socket.h"

static jmp_buf jmpbuf;

static void sigjmp(int sig)
{
	longjmp(jmpbuf, 1);
	(void) sig;
}

static void sigempty(int sig)
{
	(void) sig;
}

static void cleanup_sserv(struct sserv *sserv)
{
	char **c;

	if (sserv->empty)
		return;

	closelistener(sserv);

	free(sserv->name);
	free(sserv->protostr);
	free(sserv->path);

	if (!sserv->internal) {
		assert(sserv->argv);

		for (c = sserv->argv; *c; c++)
			free(*c);

		free(sserv->argv);
	}

	sserv->empty = 1;
}

static void usage(void)
{
	puts("Usage: inetd [-c conf] [-dh]");
	puts("");
	puts("-c <conf>   Use conf for config file [/etc/inetd.conf]");
	puts("-d          Don't fork() into background");
	puts("-h          Print this help");
	puts("");
}

int main(int argc, char *argv[])
{
	int hifd = -1, hidx = -1;
	int i, j, s;
	fd_set all, r;
	pid_t pid;
	int status;
	struct sserv *srvs;

	/* options */
	char *conf = NULL; /* /etc/inetd.conf */
	int daemonize = 1;

	while ((i = getopt(argc, argv, "c:dh")) != -1) {
		switch (i) {
		case 'c':
			conf = strdup(optarg);
			break;
		case 'd':
			daemonize = 0;
			break;
		case 'h':
		default:
			usage();
			return 1;
		}
	}

	if (!conf)
		conf = strdup("/etc/inetd.conf");

	srvs = parseconf(conf);

	if (!srvs[0].name)
		die("No valid entries found in \"%s\"\n", conf);

	free(conf);

	signal(SIGHUP, sigjmp);
	signal(SIGINT, sigjmp);
	signal(SIGQUIT, sigjmp);
	signal(SIGTERM, sigjmp);
	signal(SIGCHLD, sigempty);

	FD_ZERO(&all);

	for (i = 0; srvs[i].name; i++) {
		openlistener(&srvs[i]);
		FD_SET(srvs[i].fd, &all);
	}

	hidx = i - 1;
	hifd = srvs[hidx].fd;

	if (daemonize) {
		switch (fork()) {
		case -1:
			pdie("fork()");
		case 0:
			break;
		default:
			exit(0);
		}

		setsid();
	}

	if (setjmp(jmpbuf))
		goto end;

	for (;;) {
		r = all;

		if (select(hifd + 1, &r, NULL, NULL, NULL) == -1) {
			if (errno != EINTR)
				pdie("select()");
			goto reap;
		}

		for (i = 0; i <= hidx; i++) {
			if (!FD_ISSET(srvs[i].fd, &r))
				continue;

			if (srvs[i].wait) {
				s = srvs[i].fd;
				goto noaccept;
			}

			if ((s = accept(srvs[i].fd, NULL, NULL)) == -1)
				pdie("accept(fd %d)", srvs[i].fd);

		noaccept:
			switch (pid = fork()) {
			case -1:
				pdie("fork()");
			case 0:
				for (j = 0; j <= hidx; j++)
					if (srvs[j].fd != s)
						close(srvs[j].fd);

				close(0); close(1); close(2);
				dup(s); dup(s); dup(s);

				close(s);

				if (srvs[i].uid != geteuid())
					setuid(srvs[i].uid);
				if (srvs[i].gid != getegid())
					setgid(srvs[i].gid);

				if (srvs[i].internal) {
					switch (srvs[i].port) {
					case IPPORT_ECHO:
						builtin_echo(srvs[i].proto);
						break;
					case IPPORT_DISCARD:
						builtin_discard(srvs[i].proto);
						break;
					case IPPORT_DAYTIME:
						builtin_daytime(srvs[i].proto);
						break;
					case IPPORT_CHARGEN:
						builtin_chargen(srvs[i].proto);
						break;
					default:
						assert(0);
					}

					return 0;
				}

				execv(srvs[i].path, srvs[i].argv);

				/* uhoh */
				abort();
			default:
				srvs[i].pid = pid;

				if (srvs[i].wait)
					FD_CLR(srvs[i].fd, &all);
				else
					close(s);
				break;
			}
		}

	reap:
		while ((pid = waitpid(-1, &status, WNOHANG)) > 0)
			for (i = 0; i <= hidx; i++) {
				if (srvs[i].pid != pid)
					continue;

				srvs[i].pid = 0;

				/* Only children who exited normally or by not
				 * catching a signal (maybe abort) interest us
				 */
				if (!WIFEXITED(status) && !WIFSIGNALED(status))
					continue;

				if (srvs[i].wait)
					FD_SET(srvs[i].fd, &all);

				if (WIFSIGNALED(status) &&
				    WTERMSIG(status) == SIGABRT)
					srvs[i].aborted++;

				/* If this service keeps aborting then
				 * something is really wrong.  Take it
				 * off of our list and dispose of it.
				 */
				if (srvs[i].aborted > 1) {
					FD_CLR(srvs[i].fd, &all);
					cleanup_sserv(&srvs[i]);
				}
			}
	}

end:
	while ((pid = waitpid(-1, NULL, WNOHANG)) > 0);

	for (i = 0; i <= hidx; i++)
		cleanup_sserv(&srvs[i]);

	return 0;
}

