/* finetd, an Internet Superserver
 *
 * Supports IPv4, IPv6 and UNIX domain sockets
 *
 ** Kris Katterjohn <katterjohn@gmail.com>
 ** http://finetd.sourceforge.net
 *
 * Copyright (c) 2008 Kris Katterjohn
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

/* $Id: socket.c 10 2008-02-22 03:13:11Z kjak $ */
/* https://finetd.svn.sourceforge.net/svnroot/finetd */

#include <unistd.h>
#include <stdio.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <sys/un.h>
#include <arpa/inet.h>
#include <string.h>
#include <assert.h>

#include "finetd.h"
#include "error.h"

static void reuseaddr(int s)
{
	int one = 1;

	if (setsockopt(s, SOL_SOCKET, SO_REUSEADDR, &one, sizeof one) == -1)
		perror("setsockopt(SO_REUSEADDR)");
}

/* Does IPv4-specific functions for new socket */
static void _open_spec_in(struct sserv *sserv)
{
	struct sockaddr_in sin;

	reuseaddr(sserv->fd);

	memset(&sin, 0, sizeof sin);
	sin.sin_addr.s_addr = htonl(INADDR_ANY);
	sin.sin_port = htons(sserv->port);
	sin.sin_family = sserv->family;

	if (bind(sserv->fd, (SA *) &sin, sizeof sin) == -1)
		pdie("bind(port %d)", sserv->port);
}

/* Does IPv4-specific functions for closing socket */
static void _close_spec_in(struct sserv *sserv)
{
	(void) sserv;
}

/* Does IPv6-specific functions for new socket */
static void _open_spec_in6(struct sserv *sserv)
{
	struct sockaddr_in6 sin6;

	reuseaddr(sserv->fd);

	memset(&sin6, 0, sizeof sin6);
	sin6.sin6_addr = in6addr_any;
	sin6.sin6_port = htons(sserv->port);
	sin6.sin6_family = sserv->family;

	if (bind(sserv->fd, (SA *) &sin6, sizeof sin6) == -1)
		pdie("bind(port %d)", sserv->port);
}

/* Does IPv6-specific functions for closing socket */
static void _close_spec_in6(struct sserv *sserv)
{
	(void) sserv;
}

/* Does UNIX-specific functions for new socket */
static void _open_spec_un(struct sserv *sserv)
{
	struct sockaddr_un sun;

	unlink(sserv->name);

	memset(&sun, 0, sizeof sun);
	sun.sun_family = sserv->family;
	strncpy(sun.sun_path, sserv->name, sizeof sun.sun_path);

	if (bind(sserv->fd, (SA *) &sun, sizeof sun) == -1)
		pdie("bind(path \"%s\")", sserv->name);

	if (chown(sserv->name, sserv->un.uid, sserv->un.gid) == -1)
		pdie("chown(\"%s\", %d, %d)", sserv->name, sserv->un.uid, sserv->un.gid);

	if (chmod(sserv->name, sserv->un.mode) == -1)
		pdie("chmod(\"%s\", %d)", sserv->name, sserv->un.mode);
}

/* Does UNIX-specific functions for closing socket */
static void _close_spec_un(struct sserv *sserv)
{
	unlink(sserv->name);
}

/* Opens and sets up a new listening socket */
void openlistener(struct sserv *sserv)
{
	if ((sserv->fd = socket(sserv->family, sserv->type, sserv->proto)) == -1)
		pdie("socket(%d, %d, %d)", sserv->family, sserv->type, sserv->proto);

	switch (sserv->family) {
	case PF_INET:
		_open_spec_in(sserv);
		break;
	case PF_INET6:
		_open_spec_in6(sserv);
		break;
	case PF_UNIX:
		_open_spec_un(sserv);
		break;
	default:
		assert(0);
	}

	if (sserv->proto != IPPROTO_UDP)
		if (listen(sserv->fd, sserv->listenq) == -1)
			pdie("listen(%d, %d)", sserv->fd, sserv->listenq);
}

/* Closes a listening socket */
void closelistener(struct sserv *sserv)
{
	switch (sserv->family) {
	case PF_INET:
		_close_spec_in(sserv);
		break;
	case PF_INET6:
		_close_spec_in6(sserv);
		break;
	case PF_UNIX:
		_close_spec_un(sserv);
		break;
	default:
		assert(0);
	}

	close(sserv->fd);
}

